import copy
from collections import OrderedDict

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import hstack
from astropy.table import QTable

from exorad.log import Logger
from exorad.models.optics.opticalElement import OpticalElement
from exorad.models.signal import Radiance
from exorad.utils.diffuse_light_propagation import convolve_with_slit
from exorad.utils.diffuse_light_propagation import integrate_light
from exorad.utils.diffuse_light_propagation import prepare
from exorad.utils.exolib import planck
from exorad.utils.passVal import PassVal


def surface_radiance(wl, T, emissivity):
    """
    Spectral radiance from optics [power per unit surface, solid angle and bandwidth]

    Parameters
    __________
    wl : array_like
        wavelength array in unit of length
    T : scalar
        temperature of optical elements in [K]
    emissivity: array_like
        emissivity of optical surfaces

    Returns
    -------
    I : ndarray
        Spectral radiance. `I` will have the same shape of array `wl`

    """
    wl_ = wl.to(u.micron) if hasattr(wl, "unit") else wl
    T_ = T.to(u.K) if hasattr(T, "unit") else T
    try:
        I = planck(wl_, T_).to(u.W / (u.m**2 * u.micron * u.sr))
    except AttributeError:
        I = np.zeros_like(wl_) * u.W / u.m**2 / u.sr / u.micron
    radiance = Radiance(wl_grid=wl, data=emissivity * I)

    return radiance


class InstRadiance(Radiance):
    """
    Handler for instrument radiance

    Attributes
    ----------
    position: str
        tells the position of the optical element in the optical path. Default is 'path'
    """

    position = "path"
    slit = None


class OpticalPath(Logger):
    """
    Handler for a instrument diffuse light

    Parameters
    -----------
    description: dict
        optic description
    wl: array
        quantity array for wavelength grid

    Attributes
    ----------
    optical_element_dict_: dict
        dictionary of OpticalElement classes
    radiance_dict: dict
        dictionary of InstrRadiance classes
    radiance_table: QTable
        table of optical elements radiance
    transmission_table: QTable
        table of optical elements transmissions
    signal_table: QTable
        table of optical elements signals
    max_signal_per_pixel: QTable
        table of optical elements max signal per pixel
    slit_width: Quantity
        slit width measurement, if a slit is in the optical path. Default is None

    Methods
    -------
    prepend_optical_elements(optical_element_dict):
        updates the class optical_element_dict putting the input dictionary at the top of the existing one.
    chain():
        it concatenates the optical elements, producing the radiance_table and radiance_dictionary
    build_transmission_table():
        produces the transmission table for the optical path
    compute_signal( ch_table, ch_built_instr):
        produce the telescope self emission signal for the channel and return the channel table updated

    Examples
    -------
        >>> telescope = OpticalPath(wl=self.wl_grid, description=options)
        >>> spec = OpticalPath(wl=wl_grid, description=options['channel']['Spec'])
        >>> spec.prepend_optical_elements(telescope.optical_element_dict)
        >>> spec.build_transmission_table()
        >>> spec = spec.chain()
    """

    def __init__(self, description, wl):
        super().__init__()
        self.description = description
        self.opt = description["optics"]
        self.radiance_dict = OrderedDict()
        self.radiance_table = QTable()
        self.signal_table = QTable()
        self.max_signal_per_pixel = QTable()
        self.wl = self._wl_grid_refinement(wl)
        self.optical_element_dict = self._prepare_elements()
        self.transmission_table = QTable()
        self.slit_width = None

    def build_transmission_table(self):
        self.info("building transmission table")
        self.transmission_table["Wavelength"] = self.wl
        total_transmission = np.ones(self.wl.size)
        for el in self.optical_element_dict:
            self.transmission_table[el] = self.optical_element_dict[
                el
            ].transmission
            total_transmission *= self.optical_element_dict[el].transmission
        self.transmission_table["total"] = copy.deepcopy(total_transmission)
        self.debug("transmission table : {}".format(self.transmission_table))
        return self.transmission_table

    def _wl_grid_refinement(self, wl):
        """
        if the input wavelength list include only one element, it produce a wl_grid from the minimum wl
        investigated by the detector and the cut off
        """
        if len(wl) == 1:
            wl_min = self.description["detector"]["wl_min"]["value"].to(u.um)
            cut_off = self.description["detector"]["cut_off"]["value"].to(u.um)
            out_wl = (
                np.logspace(
                    np.log10(wl_min.value),
                    np.log10(cut_off.value),
                    PassVal.working_R,
                )
                * u.um
            )
            self.debug(
                "wl grid of a single value found. Instead we use : {}".format(
                    out_wl
                )
            )
        else:
            out_wl = wl
            self.debug("selected wl grid : {}".format(wl))
        return out_wl

    def prepend_optical_elements(self, optical_element_dict):
        self.optical_element_dict = OrderedDict(
            list(optical_element_dict.items())
            + list(self.optical_element_dict.items())
        )

    def _prepare_elements(self):
        wl = self.wl
        out = OrderedDict()
        try:
            opt_el = self.opt["opticalElement"]
        except KeyError:
            opt_el = {"value": "noElement", "type": {"value": "surface"}}
        if isinstance(opt_el, OrderedDict):
            for el in opt_el:
                self.debug("preparing {}".format(el))
                out[el] = OpticalElement(opt_el[el], wl)
        else:
            out[opt_el["value"]] = OpticalElement(opt_el, wl)
        return out

    def chain(self):
        wl = self.wl
        self.radiance_table["Wavelength"] = wl
        opt_el = self.optical_element_dict
        opt_list = list(opt_el.keys())
        self.debug("optics list : {}".format(opt_list))
        for i, k in enumerate(opt_list):
            el = opt_el[k]
            self.debug("propagating {}".format(el.name))
            out_radiance = InstRadiance(wl_grid=wl, data=np.zeros(wl.size))
            out_radiance.position = el.position
            if el.temperature is not None:
                out_radiance.data = surface_radiance(
                    el.wl, el.temperature, el.emissivity
                ).data
            else:
                self.debug("skipped because of missing temperature")
                continue
            for other_el in opt_list[i + 1 :]:
                self.debug("passing through {}".format(other_el))
                out_radiance.data *= opt_el[other_el].transmission
                if opt_el[other_el].type == "slit":
                    out_radiance.slit = True
                    self.slit_width = opt_el[other_el].description["width"][
                        "value"
                    ]
            # try:
            #     if opt_el[opt_list[i + 1]].type == 'detector box':
            #         out_radiance.position = 'optics box'
            # except IndexError:
            #     out_radiance.position = 'detector'
            self.radiance_dict[el.name] = copy.deepcopy(out_radiance)
            self.radiance_table[el.name] = copy.deepcopy(out_radiance.data)
            self.debug("final radiance : {}".format(out_radiance.data))
        return self.radiance_dict

    def compute_signal(self, ch_table, ch_built_instr):
        (
            total_max_signal,
            total_signal,
            wl_table,
            A,
            qe,
            omega_pix,
            _,
        ) = prepare(ch_table, ch_built_instr, self.description)
        for item in self.radiance_dict:
            self.debug("computing signal for {}".format(item))
            rad = copy.deepcopy(self.radiance_dict[item])
            qe.spectral_rebin(rad.wl_grid)
            if rad.slit and "slit_width" in ch_built_instr:
                max_signal_per_pix, signal = convolve_with_slit(
                    self.description,
                    ch_built_instr,
                    A,
                    ch_table,
                    omega_pix,
                    qe,
                    rad,
                )
            else:
                self.debug("no slit found")
                if rad.position == "detector":
                    self.debug("this is the detector box")
                    rad.data *= (
                        A
                        * np.pi
                        * u.sr
                        * qe.data
                        * (qe.wl_grid / const.c / const.h).to(1.0 / u.W / u.s)
                        * u.count
                    )
                elif rad.position == "optics box":
                    self.debug("this is the optics box")
                    rad.data *= (
                        A
                        * (np.pi * u.sr - omega_pix)
                        * qe.data
                        * (qe.wl_grid / const.c / const.h).to(1.0 / u.W / u.s)
                        * u.count
                    )
                else:
                    rad.data *= (
                        omega_pix
                        * A
                        * qe.data
                        * (qe.wl_grid / const.c / const.h).to(1.0 / u.W / u.s)
                        * u.count
                    )
                max_signal_per_pix, signal = integrate_light(
                    rad, rad.wl_grid, ch_built_instr
                )

            self.signal_table["{} signal".format(item)] = signal
            self.max_signal_per_pixel[item] = max_signal_per_pix

            self.debug(
                "signal : {}".format(
                    self.signal_table["{} signal".format(item)]
                )
            )
            total_signal += self.signal_table["{} signal".format(item)]
            total_max_signal += self.max_signal_per_pixel[item]
        out = QTable()
        out["instrument_signal"] = total_signal
        out["instrument_MaxSignal_inPixel"] = total_max_signal

        return hstack(
            [
                ch_table,
                out["instrument_signal", "instrument_MaxSignal_inPixel"],
            ]
        )



# Singleton Meta-Class method
# https://refactoring.guru/design-patterns/singleton/python/example


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class PassValInit(metaclass=SingletonMeta):
    """
    Class used to propagate values through the code.
    """

    _working_R = 6000

    @property
    def working_R(self):
        return self._working_R

    @working_R.setter
    def working_R(self, val):
        self._working_R = val


PassVal = PassValInit()



from scipy.interpolate import interp1d

from exorad.log import Logger
from exorad.models.signal import Radiance


class SkyFilter(Radiance):
    """
    Handler for sky filter

    Attributes
    ----------
    transmission: array
        contains the transmission information sampled at the radiance wl grid
    """

    transmission = None


class SkyForeground(Logger):
    def __init__(self, wl, description):
        super().__init__()
        self.wl = wl
        self.name = description["value"]
        self.description = description
        self.skyFilter = self._get_radiance()

    def _get_radiance(self):
        skyFilter_data = self.description["data"]

        em_func = interp1d(
            skyFilter_data["Wavelength"].to(self.wl.unit),
            skyFilter_data["Radiance"],
            assume_sorted=False,
            fill_value=0.0,
            bounds_error=False,
        )
        skyFilter = SkyFilter(
            self.wl, em_func(self.wl) * skyFilter_data["Radiance"].unit
        )
        self.debug("radiance : {}".format(skyFilter.data))

        tr_func = interp1d(
            skyFilter_data["Wavelength"].to(self.wl.unit),
            skyFilter_data["Transmission"],
            assume_sorted=False,
            fill_value=1.0,
            bounds_error=False,
        )
        skyFilter.transmission = tr_func(self.wl)
        self.debug("transmission : {}".format(skyFilter.transmission))

        return skyFilter



import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import QTable

from .instrument import Instrument
from exorad.utils.exolib import binnedPSF
from exorad.utils.exolib import find_aperture_radius
from exorad.utils.exolib import paosPSF
from exorad.utils.exolib import pixel_based_psf


class Photometer(Instrument):
    """
    Photometer instrument class.
    """

    def _wavelength_table(self):
        wl_c = 0.5 * (
            self.description["wl_min"]["value"]
            + self.description["wl_max"]["value"]
        )
        bandwidth = (
            self.description["wl_max"]["value"]
            - self.description["wl_min"]["value"]
        )
        left_bin_edge = wl_c - 0.5 * bandwidth
        right_bin_edge = wl_c + 0.5 * bandwidth
        self.table["chName"] = [self.name]
        self.table["Wavelength"] = [wl_c.value] * wl_c.unit
        self.table["Bandwidth"] = [bandwidth.value] * bandwidth.unit
        self.table["LeftBinEdge"] = [left_bin_edge.value] * left_bin_edge.unit
        self.table["RightBinEdge"] = [
            right_bin_edge.value
        ] * right_bin_edge.unit
        self.debug("wavelength table: \n{}".format(self.table))

    def builder(self):
        self.info("building {}".format(self.name))
        self._wavelength_table()
        # self.table['TR'], transmission_data = self._get_transmission()
        # self._add_data_to_built('transmission_data', transmission_data.to_dict())

        self.table["QE"], qe_data = self._get_qe()
        self._add_data_to_built("qe_data", qe_data.to_dict())

        psfFilename = None
        psf_format = None

        if "PSF" in self.description.keys():
            self.debug("PSF found")
            psfFilename = self.description["PSF"]["value"]
            if "format" in self.description["PSF"].keys():
                psf_format = self.description["PSF"]["format"]["value"]

        if psf_format == "pixel_based":
            # If psf is pixel based it simply loads the image
            prf, pixel_rf, extent = pixel_based_psf(
                self.table["Wavelength"],
                self.description["detector"]["delta_pix"]["value"],
                filename=psfFilename,
            )
        elif psf_format == "paos":

            prf, pixel_rf, extent = paosPSF(
                wl=self.table["Wavelength"],
                delta_pix=self.description["detector"]["delta_pix"]["value"],
                filename=psfFilename,
            )

        else:
            prf, pixel_rf, extent = binnedPSF(
                self.description["Fnum_x"]["value"],
                self.description["Fnum_y"]["value"],
                self.table["Wavelength"],
                self.description["detector"]["delta_pix"]["value"],
                filename=psfFilename,
            )

        self._add_data_to_built("PRF", prf)
        self._add_data_to_built("pixelRF", pixel_rf)
        self._add_data_to_built("extent", extent)

        window_size_px = self._windows_size(prf, psf_format)
        self.debug("window size : {}".format(window_size_px))
        self._add_data_to_built("window_size_px", window_size_px)
        self.table["WindowSize"] = window_size_px

        self._add_data_to_built("table", self.table)

    def _windows_size(self, prf, psf_format):
        """estimates the windows size starting from the PSF"""
        if "radius" in self.description["aperture"].keys():
            aper_radius = self.description["aperture"]["radius"]["value"]
            if psf_format == "pixel_based":
                window_size_px = [aper_radius**2]
            else:
                window_size_px = (
                    aper_radius**2
                    * self.description["Fnum_x"]["value"]
                    * self.table["Wavelength"]
                    * self.description["Fnum_y"]["value"]
                    * self.table["Wavelength"]
                    / self.description["detector"]["delta_pix"]["value"] ** 2
                )
        elif "EnE" in self.description["aperture"].keys():
            if psf_format == "pixel_based":
                aper_radius = find_aperture_radius(
                    prf,
                    self.description["aperture"]["EnE"]["value"],
                    1,
                    1,
                    1 * u.micron,
                )
                window_size_px = [np.pi * aper_radius**2 * u.Unit("")]
            else:
                aper_radius = find_aperture_radius(
                    prf,
                    self.description["aperture"]["EnE"]["value"],
                    self.description["Fnum_x"]["value"],
                    self.description["Fnum_y"]["value"],
                    self.table["Wavelength"],
                )
                window_size_px = (
                    np.pi
                    * aper_radius**2
                    * self.description["Fnum_x"]["value"]
                    * self.table["Wavelength"]
                    * self.description["Fnum_y"]["value"]
                    * self.table["Wavelength"]
                    / self.description["detector"]["delta_pix"]["value"] ** 2
                )
        return window_size_px

    def propagate_target(self, target):
        out = QTable()
        wl = target.star.sed.wl_grid
        qe, transmission, wave_window = self._get_efficiency(wl, target)
        if "sky TR" in self.table.keys():
            out["foreground_transmission"] = self.table["sky TR"]

        if self.payload["optics"]["ForceChannelWlEdge"]["value"]:
            self.debug("force channel wl edge enabled")
            idx = np.logical_or(
                wl
                < self.description["wl_min"]["value"].to(
                    self.table["Wavelength"].unit
                ),
                wl
                > self.description["wl_max"]["value"].to(
                    self.table["Wavelength"].unit
                ),
            )
            transmission[idx] = wave_window[idx] = 0.0

        star_flux = np.trapz(
            wave_window * target.star.sed.data, x=target.star.sed.wl_grid
        ).to(u.W / u.m**2)
        out["starFlux"] = [star_flux.value] * star_flux.unit
        self.debug("star flux : {}".format(out["starFlux"]))

        star_signal = (
            self.payload["optics"]["Atel"]["value"]
            * np.trapz(
                qe
                * target.star.sed.data
                * transmission
                * target.star.sed.wl_grid.to(u.m)
                / const.c
                / const.h,
                x=target.star.sed.wl_grid,
            ).si
            * u.count
        )
        out["starSignal"] = [star_signal.value] * star_signal.unit
        self.debug("star signal : {}".format(out["starSignal"]))

        if "apertureCorrection" in self.description["aperture"].keys():
            star_signal_aperture = (
                star_signal
                * self.description["aperture"]["apertureCorrection"]["value"]
            )
        elif "EnE" in self.description["aperture"].keys():
            star_signal_aperture = (
                star_signal * self.description["aperture"]["EnE"]["value"]
            )
        else:
            self.debug("aperture correction not found")
            star_signal_aperture = star_signal

        out["star_signal_inAperture"] = [
            star_signal_aperture.value
        ] * star_signal_aperture.unit
        self.debug(
            "star signal in aperture : {}".format(
                out["star_signal_inAperture"]
            )
        )

        star_signal_in_pixel = self.built_instr["PRF"].max() * star_signal
        out["star_MaxSignal_inPixel"] = [
            star_signal_in_pixel.value
        ] * star_signal_in_pixel.unit
        self.debug(
            "star signal in pixel MAX : {}".format(
                out["star_signal_inAperture"]
            )
        )
        return out



from .foregroundHandler import EstimateForeground
from .foregroundHandler import EstimateForegrounds
from .foregroundHandler import EstimateZodi
from .instrumentHandler import BuildChannels
from .instrumentHandler import BuildInstrument
from .instrumentHandler import GetChannelList
from .instrumentHandler import LoadPayload
from .instrumentHandler import MergeChannelsOutput
from .instrumentHandler import PreparePayload
from .loadOptions import LoadOptions
from .loadSource import LoadSource
from .noiseHandler import EstimateNoise
from .noiseHandler import EstimateNoiseInChannel
from .propagateLight import PropagateForegroundLight
from .propagateLight import PropagateTargetLight
from .targetHandler import EstimateMaxSignal
from .targetHandler import LoadTargetList
from .targetHandler import ObserveTarget
from .targetHandler import ObserveTargetlist
from .targetHandler import PrepareTarget
from .targetHandler import UpdateTargetTable



import re

import numpy as np
import xlrd

xlrd.xlsx.ensure_elementtree_imported(False, None)
xlrd.xlsx.Element_has_iter = True

from astropy import units as u
from astropy.io import ascii

from exorad.log import Logger
from exorad.models.target import Target

compactString = lambda string: string.replace(" ", "").replace("-", "").lower()
stripUnitString = lambda string: string.replace("[", "").replace("]", "")


class BaseTargetList(Logger):
    """
    Target list base class
    """

    def __init__(self):
        self.set_log_name()

        self.read_data()
        self._targets = self.create_target_list()

    def star_keys(self):
        raise NotImplementedError

    def star_data(self):
        raise NotImplementedError

    def read_data(self):
        raise NotImplementedError

    def create_target_list(self):

        star_keys = self.star_keys()

        star_data = self.star_data()

        # print(planet_data)
        # print(type(planet_data))
        # quit()
        planet_data = None
        planet_keys = None
        try:
            planet_keys = self.planet_keys()
            planet_data = self.planet_data()
        except:
            pass

        target_list = []
        if planet_data:
            for i, (star, planet) in enumerate(zip(star_data, planet_data)):
                target = Target()
                target.planet = Target()
                target.star = Target()

                planet_dict = dict(zip(planet_keys, planet))

                target.planet.__dict__.update(planet_dict)

                star_dict = dict(zip(star_keys, star))

                target.star.__dict__.update(star_dict)
                target.name = target.planet.name
                target.id = i

                target_list.append(target)
        else:
            for i, star in enumerate(star_data):
                target = Target()
                target.star = Target()
                target.id = i

                star_dict = dict(zip(star_keys, star))

                target.star.__dict__.update(star_dict)
                target.name = target.star.name

                target_list.append(target)
        return target_list

    @property
    def target(self):
        return self._targets

    def searchTarget(self, name):
        # method inspired from similar functionality in ExoData
        searchName = compactString(name)
        returnList = []

        for target in self.target:
            if re.search(
                searchName, compactString(str(target.star.name))
            ) or re.search(searchName, compactString(str(target.planet.name))):
                returnList.append(target)

        return returnList


class XLXSTargetList(BaseTargetList):
    """
    It parses a .xlsx file into a :class:`BaseTargetList` class
    """

    star_data_columns = (1, 6)
    planet_data_columns = (8, 18)
    number_to_be_observed_column = 20
    data_row0 = 4
    units_row = 3

    def __init__(self, filename):
        self.filename = filename

        super().__init__()

    def __parseData__(self, col_range):
        Sheet = self.tmpSheet
        keys = Sheet.row_values(2)[col_range[0] : col_range[1] + 1]
        keys[0] = "name"
        obj = {}

        for key, col in zip(keys, list(range(col_range[0], col_range[1] + 1))):
            str_unit = Sheet.cell_value(rowx=self.units_row, colx=col)
            if len(str_unit) > 0:
                try:
                    dim = u.Unit(stripUnitString(str_unit))
                except ValueError:
                    self.warning("Unrecognised physical units")
                    dim = 1
            else:
                dim = 1

            obj[key] = Sheet.col_values(col)[self.data_row0 :] * dim

        return obj

    def planet_keys(self):
        return self.planet.keys()

    def star_keys(self):
        return self.star.keys()

    def planet_data(self):
        planet_list = list(self.planet.values())
        return list(map(list, zip(*planet_list)))

    def star_data(self):
        star_list = list(self.star.values())
        return list(map(list, zip(*star_list)))

    def read_data(self):

        xl_workbook = xlrd.open_workbook(self.filename)
        self.tmpSheet = xl_workbook.sheet_by_name("Sheet1")

        self.planet = self.__parseData__(col_range=self.planet_data_columns)
        self.planet["Nobs"] = self.tmpSheet.col_values(
            self.number_to_be_observed_column
        )[self.data_row0 :]

        self.star = self.__parseData__(col_range=self.star_data_columns)


class CSVTargetList(BaseTargetList):
    """
    It parses a csv file into a :class:`BaseTargetList` class
    """

    def __init__(self, filename):
        self.filename = filename
        super().__init__()

    def read_data(self):
        self.tmpTab = ascii.read(self.filename, format="csv")

    def star_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "star" in k]
        return [k.split(" ")[1] for k in s_col]

    def star_data(self):
        star_k = [k for k in self.tmpTab.keys() if "star" in k]
        units = []
        for n, k in enumerate(star_k):
            if len(k.split(" ")) == 3:
                un = k.split(" ")[2]
                units.append(1 * u.Unit(stripUnitString(un)))
            else:
                units.append(None)
        dat = []
        for i in range(len(self.tmpTab[star_k])):
            val = list(self.tmpTab[star_k][i])

            for n in range(len(val)):
                if units[n] is not None:
                    val[n] = val[n] * units[n]
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat

    def planet_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "planet" in k]
        return [k.split(" ")[1] for k in s_col]

    def planet_data(self):
        planet_k = [k for k in self.tmpTab.keys() if "planet" in k]
        units = []
        for n, k in enumerate(planet_k):
            if len(k.split(" ")) == 3:
                un = k.split(" ")[2]
                units.append(1 * u.Unit(stripUnitString(un)))
            else:
                units.append(None)
        dat = []
        for i in range(len(self.tmpTab[planet_k])):
            val = list(self.tmpTab[planet_k][i])

            for n in range(len(val)):
                if units[n] is not None:
                    val[n] = val[n] * units[n]
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat


class CSVTargetListMRS(BaseTargetList):
    """
    It parses a csv file into a :class:`BaseTargetList` class
    """

    def __init__(self, filename):
        self.filename = filename
        super().__init__()

    def read_data(self):
        self.tmpTab = ascii.read(self.filename, format="csv")

    def star_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "star" in k.lower()]
        s_col = [
            k.split(" ", 1)[1].lower()
            for k in s_col
            if "error" not in k.lower()
        ]
        return [
            self.check_keys(k[0 : k.find("[") - 1])
            if "[" in k
            else self.check_keys(k)
            for k in s_col
        ]

    def star_data(self):
        star_k = [
            k
            for k in self.tmpTab.keys()
            if "star" in k.lower() and "error" not in k.lower()
        ]
        units = []
        for n, k in enumerate(star_k):
            if "[" in k:
                units.append(
                    1
                    * u.Unit(
                        self.check_units(k[k.find("[") + 1 : k.find("]")])
                    )
                )
            else:
                units.append(None)
        dat = []
        for i in range(len(self.tmpTab[star_k])):
            val = list(self.tmpTab[star_k][i])
            for n in range(len(val)):
                if isinstance(val[n], str):
                    if (
                        "," in val[n]
                    ):  # for some reason, comma are used instead of dots sometimes
                        val[n] = val[n].replace(",", ".")
                    try:
                        float(val[n])
                    except ValueError:
                        val[n]
                if units[n] is not None:
                    val[n] = val[n] * units[n]
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat

    def planet_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "planet" in k.lower()]
        s_col = [
            k.split(" ", 1)[1].lower()
            for k in s_col
            if "error" not in k.lower()
        ]
        return [
            self.check_keys(k[0 : k.find("[") - 1])
            if "[" in k
            else self.check_keys(k)
            for k in s_col
        ]

    def planet_data(self):
        planet_k = [
            k
            for k in self.tmpTab.keys()
            if "planet" in k.lower() and "error" not in k.lower()
        ]
        units = []
        for n, k in enumerate(planet_k):
            if "[" in k:
                units.append(
                    1
                    * u.Unit(
                        self.check_units(k[k.find("[") + 1 : k.find("]")])
                    )
                )
            else:
                units.append(None)
        dat = []
        for i in range(len(self.tmpTab[planet_k])):
            val = list(self.tmpTab[planet_k][i])
            for n in range(len(val)):
                if isinstance(val[n], str):
                    if (
                        "," in val[n]
                    ):  # for some reason, comma are used instead of dots sometimes
                        val[n] = val[n].replace(",", ".")
                    try:
                        float(val[n])
                    except ValueError:
                        val[n]
                if units[n] is not None:
                    val[n] = val[n] * units[n]
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat

    @staticmethod
    def check_units(unit):
        if unit == "Rs":
            return "R_sun"
        if unit == "Ms":
            return "M_sun"
        if unit == "Re":
            return "R_earth"
        if unit == "Me":
            return "M_earth"
        if unit == "Rj":
            return "R_jupiter"
        if unit == "Mj":
            return "M_jupiter"
        if unit == "days":
            return "day"
        else:
            return unit

    @staticmethod
    def check_keys(key):
        if key == "mass":
            return "M"
        if key == "radius":
            return "R"
        if key == "temperature":
            return "Teff"
        if key == "distance":
            return "D"
        if key == "period":
            return "P"
        if key == "molecular weight":
            return "mmw"
        if key == "semi-major axis":
            return "a"
        if key == "impact parameter":
            return "b"
        if key == "heat redistribution factor":
            return "Gamma"
        if key == "transit duration":
            return "T14"
        else:
            return key


class OldExcelTargetList(Logger):
    """
    It parses an .xls file into a :class:`BaseTargetList` class
    """

    star_data_columns = (1, 6)
    number_to_be_observed_column = 20
    data_row0 = 4
    units_row = 3

    def __init__(self, filename):

        self.set_log_name()

        if filename.endswith(".xlsx"):
            xl_workbook = xlrd.open_workbook(filename)

            tmpSheet = xl_workbook.sheet_by_name("Sheet1")
            tmpSheet = tmpSheet

            star = self.__parseData__(
                tmpSheet, col_range=self.star_data_columns
            )

            key0 = list(star.keys())[0]
            n_targets = len(star[key0])

            self.target = [Target() for k in range(n_targets)]
            for k, target in enumerate(self.target):
                setattr(target, "star", Target())
                for key in list(star.keys()):
                    setattr(target.star, key, star[key][k])
        else:
            self.error("Wrong target list format")
            raise OSError("Wrong target list format")

    def __parseData__(self, Sheet, col_range):
        keys = Sheet.row_values(2)[col_range[0] : col_range[1] + 1]
        keys[0] = "name"
        obj = {}

        for key, col in zip(keys, list(range(col_range[0], col_range[1] + 1))):
            str_unit = Sheet.cell_value(rowx=self.units_row, colx=col)
            if len(str_unit) > 0:
                try:
                    dim = u.Unit(stripUnitString(str_unit))
                except ValueError:
                    self.warning("Unrecognised physical units")
                    dim = 1
            else:
                dim = 1

            obj[key] = Sheet.col_values(col)[self.data_row0 :] * dim

        return obj

    def searchTarget(self, name):
        # method inspired from similar functionality in ExoData
        searchName = compactString(name)
        returnList = []

        for target in self.target:
            if re.search(
                searchName, compactString(str(target.star.name))
            ) or re.search(searchName, compactString(str(target.planet.name))):
                returnList.append(target)

        return returnList


class QTableTargetList(BaseTargetList):
    """
    It parses an :class:`astropy.table.QTable` into a :class:`BaseTargetList` class
    """

    def __init__(self, table):
        self.input_table = table
        super().__init__()

    def read_data(self):
        self.tmpTab = self.input_table

    def star_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "star" in k]
        return [k.split(" ")[1] for k in s_col]

    def star_data(self):
        star_k = [k for k in self.tmpTab.keys() if "star" in k]
        dat = []
        for i in range(len(self.tmpTab[star_k])):
            val = list(self.tmpTab[star_k][i])
            self._check_quantities(val)
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat

    def planet_keys(self):
        s_col = [k for k in self.tmpTab.keys() if "planet" in k]
        return [k.split(" ")[1] for k in s_col]

    def planet_data(self):
        star_k = [k for k in self.tmpTab.keys() if "planet" in k]
        dat = []
        for i in range(len(self.tmpTab[star_k])):
            val = list(self.tmpTab[star_k][i])
            self._check_quantities(val)
            dat.append([str(x) if isinstance(x, np.str_) else x for x in val])
        return dat

    def _check_quantities(self, val):
        for v in val:
            if not isinstance(v, np.str_):
                if not hasattr(v, "unit"):
                    self.error(
                        "Wrong target list format: Quantities are required"
                    )
                    raise TypeError(
                        "Wrong target list format: Quantities are required"
                    )



import logging
import os

import astropy.units as u
import h5py
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from astropy.io.misc.hdf5 import read_table_hdf5

from exorad.log import Logger
from exorad.log import setLogLevel

mpl_logger = logging.getLogger("matplotlib")
mpl_logger.setLevel(logging.WARNING)
cmap = matplotlib.cm.get_cmap("Set1")


class Plotter(Logger):
    """
    Plotter class. It offers a fast and easy way to produce diagnostic plots on the produced data.

    Parameters
    -----------
    input_table: Qtable
        table where to grab data and wl grid to plot.
    channels: dict
        dictionary describing the channels in the payload. Default is None
    """

    def __init__(self, input_table, channels=None, payload=None):
        self.set_log_name()
        self.inputTable = input_table
        self.channels = channels
        self.payload = payload
        self.fig = None
        self.fig_efficiency = None

    def plot_bands(self, ax, scale="log", channel_edges=True):
        """
        It plots the channels bands behind the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the bands

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        channels = set(self.inputTable["chName"])
        norm = matplotlib.colors.Normalize(vmin=0.0, vmax=len(channels))
        tick_list = []

        for k, channelName in enumerate(channels):
            wl_min = min(
                self.inputTable["LeftBinEdge"][
                    np.where(self.inputTable["chName"] == channelName)
                ]
            )
            if hasattr(wl_min, "unit"):
                wl_min = wl_min.value
            wl_max = max(
                self.inputTable["RightBinEdge"][
                    np.where(self.inputTable["chName"] == channelName)
                ]
            )
            if hasattr(wl_max, "unit"):
                wl_max = wl_max.value
            ax.axvspan(
                wl_min,
                wl_max,
                alpha=0.1,
                zorder=0,
                color=cmap(
                    norm(k),
                ),
            )
            tick_list.append(wl_min)
        tick_list.append(wl_max)
        ax.set_xscale(scale)
        if channel_edges:
            ax.set_xticks(tick_list)
            ax.get_xaxis().set_major_formatter(
                matplotlib.ticker.ScalarFormatter()
            )

        return ax

    def plot_efficiency(self, scale="log", channel_edges=True):
        """
        It produces a figure with payload efficiency over wavelength.
        The quantities reported are quantum efficiency, transmission and the photon conversion efficiency (pce)
        computed as the product of the quantum efficiency and transmission.

        Returns
        --------
        matplotlib.pyplot.figure

        Note
        ----
        The Class channels input parameter is required for this method to work.
        """
        if self.channels:
            from matplotlib.lines import Line2D
            from exorad.models.signal import Signal

            norm = matplotlib.colors.Normalize(
                vmin=0.0, vmax=len(self.channels)
            )

            fig, ax = plt.subplots(1, 1, figsize=(10, 8))
            fig.suptitle("Payload photon conversion efficiency")

            if self.channels:
                keys = ["transmission", "qe"]
                for ch in self.channels:
                    pce = None
                    for i, key in enumerate(keys):
                        data = self.channels[ch].built_instr[
                            "{}_data".format(key)
                        ]
                        sig = Signal(
                            wl_grid=data["wl_grid"]["value"]
                            * u.Unit(data["wl_grid"]["unit"]),
                            data=data["data"]["value"],
                        )
                        ax.plot(
                            sig.wl_grid,
                            sig.data,
                            color=cmap(
                                norm(i),
                            ),
                            zorder=10,
                        )
                        if not pce:
                            pce = sig
                        else:
                            sig.spectral_rebin(pce.wl_grid)
                            pce.data *= sig.data
                    ax.plot(
                        pce.wl_grid,
                        pce.data,
                        color=cmap(
                            norm(i + 1),
                        ),
                        zorder=10,
                    )

                lines, labels = [], []
                for i, key in enumerate(keys):
                    lines.append(
                        Line2D(
                            [0],
                            [0],
                            color=cmap(
                                norm(i),
                            ),
                            lw=4,
                        )
                    )
                    labels.append(key)
                lines.append(
                    Line2D(
                        [0],
                        [0],
                        color=cmap(
                            norm(i + 1),
                        ),
                        lw=4,
                    )
                )
                labels.append("pce")
            else:
                pce = self.inputTable["TR"] * self.inputTable["QE"]
                self.inputTable["pce"] = pce
                keys = ["TR", "QE", "pce"]
                self.debug("efficiency keys : {}".format(keys))
                for e in keys:
                    ax.plot(
                        self.inputTable["Wavelength"],
                        self.inputTable[e],
                        label=e,
                        zorder=10,
                    )
                    # ax.plot(self.inputTable['Wavelength'], self.inputTable[e], c='None')
            ax.grid(zorder=0)
            ax = self.plot_bands(ax, scale, channel_edges)
            ax.legend(handles=lines, labels=labels)
            ax.set_title("Photon conversion efficiency")
            ax.set_xlabel(r"Wavelength [$\mu m$]")
            ax.set_ylabel("efficiency")
            # ax.set_xscale('log')

            self.fig_efficiency = fig
            return fig

        else:
            self.error(
                "channels parameter is required for this method to work"
            )

    def plot_noise(self, ax, ylim=None, scale="log", channel_edges=True):
        """
        It plots the noise components found in the input table in the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the noises
        ylim: float
            if present, it sets the axes y bottom lim. Default is None.

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        noise_keys = [
            x for x in self.inputTable.keys() if "noise" in x or "custom" in x
        ]
        self.debug("noise keys : {}".format(noise_keys))
        for k, n in enumerate(noise_keys):
            if n == "total_noise":
                ax.plot(
                    self.inputTable["Wavelength"],
                    self.inputTable[n],
                    zorder=9,
                    lw=1,
                    c="k",
                    marker=".",
                    markersize=5,
                    label="total_noise",
                    alpha=0.8,
                )  # , c='None')
            else:
                if self.inputTable[n].unit == u.hr**0.5:
                    noise = self.inputTable[n]
                else:
                    self.debug(
                        "{} rescaled by starSignal_inAperture".format(n)
                    )
                    noise = (
                        self.inputTable[n]
                        / self.inputTable["star_signal_inAperture"]
                    )
                # ax.scatter(self.inputTable['Wavelength'], noise, label=n, zorder=10, s=5, color=palette[k])
                ax.plot(
                    self.inputTable["Wavelength"],
                    noise,
                    zorder=9,
                    lw=1,
                    alpha=0.5,
                    marker=".",
                    label=n,
                )  # color=palette[k])  # c='None')

        if not ylim:
            ax.set_ylim(1e-7)
        else:
            ax.set_ylim(ylim)
        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(
            base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12
        )
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis="y", which="minor", alpha=0.3)
        ax.grid(axis="y", which="major", alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.legend(
            prop={"size": 12},
            loc="upper left",
            ncol=3,
            bbox_to_anchor=(0.05, -0.2),
            labelspacing=1.2,
            handlelength=1,
        )
        ax.set_title("Noise Budget")
        ax.set_xlabel(r"Wavelength [$\mu m$]")
        ax.set_ylabel(r"relative noise [$\sqrt{{hr}}$]")
        ax.set_yscale("log")
        # ax.set_xscale('log')
        ax = self.plot_bands(ax, scale, channel_edges)
        return ax

    def plot_signal(self, ax, ylim=None, scale="log", channel_edges=True):
        """
        It plots the signal components found in the input table in the indicated axes.

        Parameters
        -----------
        ax: matplotlib.axes
            axes where to plot the signals
        ylim: float
            if present, it sets the axes y bottom lim. Default is None.

        Returns
        --------
        matplotlib.axes.axes

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """

        keys = [
            x
            for x in self.inputTable.keys()
            if "signal" in x and "noise" not in x
        ]
        self.debug("signal keys : {}".format(keys))
        for k, s in enumerate(keys):
            # ax.scatter(self.inputTable['Wavelength'], self.inputTable[s], label=s, zorder=10, s=5, color=palette[k])
            ax.plot(
                self.inputTable["Wavelength"],
                self.inputTable[s],
                zorder=9,
                lw=1,
                alpha=0.5,
                marker=".",
                label=s,
            )
            # color=palette[k])  # , c='None')

        if not ylim:
            ax.set_ylim(1e-3)
        else:
            ax.set_ylim(ylim)
        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(
            base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12
        )
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis="y", which="minor", alpha=0.3)
        ax.grid(axis="y", which="major", alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.legend(
            prop={"size": 12},
            loc="upper left",
            ncol=3,
            bbox_to_anchor=(0.05, -0.2),
            labelspacing=1.2,
            handlelength=1,
        )
        ax.set_title("Signals")
        ax.set_xlabel(r"Wavelength [$\mu m$]")
        ax.set_ylabel("$ct/s$")
        ax.set_yscale("log")
        # ax.set_xscale('log')
        ax = self.plot_bands(ax, scale, channel_edges)

        return ax

    def plot_table(
        self, sig_ylim=None, noise_ylim=None, scale="log", channel_edges=True
    ):
        """
        It produces a figure with signal and noise for the input table.

        Returns
        --------
        matplotlib.pyplot.figure
        (matplotlib.axes.axes, matplotlib.axes.axes)

        Note
        ----
        The Class input_table input parameter is required for this method to work.
        """
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 10))
        fig.suptitle(self.inputTable.meta["name"])
        ax1 = self.plot_signal(
            ax1, ylim=sig_ylim, scale=scale, channel_edges=channel_edges
        )
        ax2 = self.plot_noise(
            ax2, ylim=noise_ylim, scale=scale, channel_edges=channel_edges
        )
        plt.tight_layout()
        plt.subplots_adjust(top=0.9, bottom=0.22, hspace=0.7)
        self.fig = fig
        return fig, (ax1, ax2)

    def save_fig(self, name, efficiency=False):
        """
        It saves the produced figure.

        Parameters
        --------
        name: str
            figure name
        efficiency: bool
            if True it wll save the efficiency plot instead of the table plot. Default is False.
        """
        try:
            if efficiency:
                self.fig_efficiency.savefig("{}".format(name))
            else:
                self.fig.savefig("{}".format(name))

            self.info("plot saved in {}".format(name))
        except AttributeError:
            self.error(
                "the indicated figure is not available. Check if you have produced it."
            )


def main():
    import argparse
    from exorad.__version__ import __version__
    from exorad.utils.util import parse_range

    parser = argparse.ArgumentParser(
        description="ExoRad {}".format(__version__)
    )
    parser.add_argument(
        "-i",
        "--input",
        dest="input",
        type=str,
        required=True,
        help="Input h5 file to pass",
    )
    parser.add_argument(
        "-o",
        "--out",
        dest="out",
        type=str,
        default="None",
        required=True,
        help="Output directory",
    )
    parser.add_argument(
        "-n",
        "--target-number",
        dest="target_number",
        type=str,
        default="all",
        required=False,
        help="A list or range of targets to run",
    )
    parser.add_argument(
        "-t",
        "--target-name",
        dest="target_name",
        type=str,
        default="None",
        required=False,
        help="name of the target to plot",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        default=False,
        required=False,
        help="log output on screen",
        action="store_true",
    )

    args = parser.parse_args()

    logger = logging.getLogger("exorad")
    from exorad.utils.ascii_art import ascii_plot

    logger.info(ascii_plot)
    logger.info("code version {}".format(__version__))

    if args.debug:
        setLogLevel(logging.DEBUG)

    if not os.path.exists(args.out):
        os.makedirs(args.out)
        logger.info("output directory created")

    logger.info("reading {}".format(args.input))
    file = h5py.File(args.input)

    if args.target_number != "all" and args.target_name != "None":
        logger.error("you cannot use both target number and target name")
        raise ValueError

    targets_dir = file["targets"]
    targets_to_run_id = parse_range(
        args.target_number, len(targets_dir.keys())
    )
    targets_to_run = [list(targets_dir.keys())[n] for n in targets_to_run_id]

    if args.target_name != "None":
        targets_to_run = [
            target for target in targets_to_run if target == args.target_name
        ]

    for target in targets_to_run:
        target_dir = targets_dir[target]
        table_dir = target_dir["table"]
        table = read_table_hdf5(table_dir, path="table")

        plotter = Plotter(input_table=table)
        plotter.plot_table()
        plotter.save_fig(os.path.join(args.out, "{}.png".format(target)))
        plt.close()



import astropy.units as u

from exorad.log import Logger
from exorad.models.signal import Radiance
from exorad.utils.exolib import planck


class ZodiacalFrg(Logger):
    """
    Produce the zodiacal radiance.

    Parameters
    ----------
    wl: array
        wavelength grid
    description: dict
        zodiacal foreground description
    coordinates: (float, float)
        pointing coordinates as (ra, dec)

    Attributes
    -----------
    wl: array
        wavelength grid
    description: dict
        zodiacal foreground description dict
    radiance: Radiance
        zodiacal radiance

    """

    def __init__(self, wl, description, coordinates=None):
        super().__init__()
        self.wl = wl
        self.description = description
        self.radiance = self._get_radiance(coordinates)

    def _get_radiance(self, coordinates):

        fit_coord = (
            self.description["zodiacalMap"]["value"]
            if "zodiacalMap" in self.description
            else False
        )
        map_file = (
            self.description["custom_map_file"]["value"]
            if "custom_map_file" in self.description
            else None
        )
        if fit_coord and coordinates:
            try:
                A = self.zodiacal_fit_direction(coordinates, map_file)
            except (KeyError, OSError):
                self.warning("zodiacal fit failed")
                self._get_radiance(coordinates=None)

        elif "zodiacFactor" in self.description:
            self.debug(" model used for zodiacal foreground")
            A = self.description["zodiacFactor"]["value"]
            self.debug("zodiac factor : {}".format(A))
        else:
            self.warning("zodiacal description uncompleted")
            A = 0

        return self.model(A)

    def model(self, A):
        """
        returns the zodiacal Radiance for the zodiacal light model presented in Glasse et al. 2010,
        scaled by coefficient fitted over Kelsall et al. 1998 model
        """
        units = u.W / (u.m**2 * u.um * u.sr)
        zodi_emission = A * (
            3.5e-14 * planck(self.wl, 5500.0 * u.K)
            + 3.58e-8 * planck(self.wl, 270.0 * u.K)
        ).to(units)
        return Radiance(self.wl, zodi_emission)

    def zodiacal_fit_direction(self, coord, map_file=None):
        """
        returns the fitted coefficient for the zodiacal model given the target position.
        It's based on Kelsall et al. 1998 model
        """
        import os
        from astropy.io.misc.hdf5 import read_table_hdf5
        import numpy as np
        from pathlib import Path

        ra_input = coord[0]
        dec_input = coord[1]

        exorad_path = Path(__file__).parent.parent.parent.absolute()
        zodi_map_file = (
            map_file
            if map_file
            else os.path.join(exorad_path, "data/Zodi_map.hdf5")
        )

        if not os.path.isfile(zodi_map_file):
            dir_path = Path(__file__).parent.absolute()

            i = 0
            while (
                "data"
                not in [d.stem for d in Path(dir_path).iterdir() if d.is_dir()]
                and i < 10
            ):
                dir_path = dir_path.parent
                i += 1
            if i > 10:
                self.error("Zodi map file not found")
                raise OSError("Zodi map file not found")

            data_path = os.path.join(dir_path.absolute().as_posix(), "data")
            zodi_map_file = os.path.join(data_path, "Zodi_map.hdf5")

        self.debug("map data:{}".format(zodi_map_file))

        try:
            zodi_table = read_table_hdf5(zodi_map_file)
            self.debug(zodi_table)
            distance = (zodi_table["ra_icrs"] * u.deg - ra_input) ** 2 + (
                zodi_table["dec_icrs"] * u.deg - dec_input
            ) ** 2
            idx = np.argmin(distance)
            self.debug("selected line {}".format(idx))
            self.debug(zodi_table[idx])
            return zodi_table["zodi_coeff"][idx]

        except OSError:
            self.error("Zodi map file not found")
            raise OSError("Zodi map file not found")



import logging
import os
import pathlib
import unittest

import astropy.units as u
import numpy as np
from astropy import constants as cc

from exorad.log import setLogLevel

setLogLevel(logging.DEBUG)


class StarTest(unittest.TestCase):

    phoenix_stellar_model = '/usr/local/project_data/sed/'

    @unittest.skipIf(not os.path.isdir(phoenix_stellar_model),
                     'phoenix dir not found')
    def test_phoenix_star(self):
        from exorad.models.source import Star

        target = {'D': 12.975 * u.pc,
                  'T': 3016 * u.K,
                  'M': 0.15 * u.Msun,
                  'R': 0.218 * u.Rsun}
        g = (cc.G * target['M'].si / (target['R'].si) ** 2).to(u.cm / u.s ** 2)
        logG = np.log10(g.value)

        star1 = Star(self.phoenix_stellar_model,
             target['D'],
             target['T'],
             logG,
             0.0,
             target['R'],
             use_planck_spectrum=False)

    # todo test more phoenix formats
    # def test_phoenix_star_fits(self):
    #     from exorad.models.source import Star
    #
    #     from exorad.models.source import Star
    #     path = pathlib.Path(__file__).parent.absolute()
    #     test_dir = os.path.join(path, 'test_data/fits')
    #
    #     target = {'D': 12.975 * u.pc,
    #               'T': 2300 * u.K,
    #               'M': 0.15 * u.Msun,
    #               'R': 0.218 * u.Rsun}
    #     g = (cc.G * target['M'].si / (target['R'].si) ** 2).to(u.cm / u.s ** 2)
    #     logG = np.log10(g.value)
    #
    #     star1 =Star(test_dir,
    #          target['D'],
    #          target['T'],
    #          logG,
    #          0.0,
    #          target['R'],
    #          use_planck_spectrum=False)
    #
    #
    #     bb = Star('.',
    #          target['D'],
    #          target['T'],
    #          0.0,
    #          0.0,
    #          target['R'],
    #          wl_min=star1.sed.wl_grid.value.min() * u.um,
    #          wl_max=star1.sed.wl_grid.value.max() * u.um,
    #          use_planck_spectrum=True)
    #
    #     import matplotlib.pyplot as plt
    #     plt.plot(star1.sed.wl_grid, star1.sed.data, label='phoenix')
    #     plt.plot(bb.sed.wl_grid, bb.sed.data, label='planck')
    #     plt.yscale('log')
    #     plt.legend()
    #     plt.show()
    #
    # def test_phoenix_star_7format(self):
    #     from exorad.models.source import Star
    #     path = pathlib.Path(__file__).parent.absolute()
    #     test_dir = os.path.join(path, 'test_data')
    #
    #
    #     target = {'D': 12.975 * u.pc,
    #               'T': 800 * u.K,
    #               'M': 0.15 * u.Msun,
    #               'R': 0.218 * u.Rsun}
    #     g = (cc.G * target['M'].si / (target['R'].si) ** 2).to(u.cm / u.s ** 2)
    #     logG = np.log10(g.value)
    #
    #     star1 =Star(test_dir,
    #          target['D'],
    #          target['T'],
    #          logG,
    #          0.0,
    #          target['R'],
    #          use_planck_spectrum=False)
    #
    #     bb = Star('.',
    #          target['D'],
    #          target['T'],
    #          0.0,
    #          0.0,
    #          target['R'],
    #          wl_min=star1.sed.wl_grid.value.min() * u.um,
    #          wl_max=star1.sed.wl_grid.value.max() * u.um,
    #          use_planck_spectrum=True)
    #
    #     import matplotlib.pyplot as plt
    #     plt.plot(star1.sed.wl_grid, star1.sed.data, label='phoenix')
    #     plt.plot(bb.sed.wl_grid, bb.sed.data, label='planck')
    #     # plt.xlim(bb.sed.wl_grid.value.min(), bb.sed.wl_grid.value.max())
    #     plt.yscale('log')
    #     plt.legend()
    #     plt.show()

    def test_BB_star(self):
        from exorad.models.source import Star

        target = {'D': 12.975 * u.pc,
                  'T': 3016 * u.K,
                  'M': 0.15 * u.Msun,
                  'R': 0.218 * u.Rsun}

        Star('.',
             target['D'],
             target['T'],
             0.0,
             0.0,
             target['R'],
             use_planck_spectrum=True)

    def test_GenericSed(self):
        import os
        from inspect import getsourcefile
        from exorad.models.source import CustomSed
        current_folder = os.path.dirname(
            os.path.abspath(getsourcefile(lambda: 0)))
        target = {'D': 12.975 * u.pc,
                  'R': 0.218 * u.Rsun}
        CustomSed(os.path.join(os.path.dirname(current_folder),
                               'examples/customsed.csv'),
                  target['R'],
                  target['D'],
                  )



import logging
import os
import pathlib
import unittest

import matplotlib.pyplot as plt
from conf import skip_plot
from test_options import payload_file

from exorad import tasks
from exorad.log import disableLogging
from exorad.log import enableLogging
from exorad.log import setLogLevel
from exorad.utils.plotter import Plotter

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')


class PlotterTest(unittest.TestCase):
    loadOptions = tasks.LoadOptions()
    buildChannels = tasks.BuildChannels()
    buildInstrument = tasks.BuildInstrument()
    loadPayload = tasks.LoadPayload()
    loadTargetList = tasks.LoadTargetList()
    loadSource = tasks.loadSource.LoadSource()
    mergeChannelsOutput = tasks.MergeChannelsOutput()
    observeTargetList = tasks.ObserveTargetlist()

    disableLogging()
    payload = loadOptions(filename=payload_file())
    wl_min, wl_max = payload['common']['wl_min']['value'], \
                     payload['common']['wl_max']['value']
    channels = buildChannels(payload=payload, write=False, output=None)
    table = mergeChannelsOutput(channels=channels)

    targets = loadTargetList(
        target_list=os.path.join(data_dir, 'test_target.csv'))
    targets = observeTargetList(targets=targets.target, payload=payload,
                                channels=channels, wl_range=(wl_min, wl_max),
                                plot=False, out_dir=None)
    enableLogging()

    @unittest.skipIf(skip_plot, "This test only produces plots")
    def test_efficiency(self):
        setLogLevel(logging.DEBUG)

        plotter = Plotter(channels=self.channels, input_table=self.table)
        plotter.plot_efficiency()
        plt.show()

    @unittest.skipIf(skip_plot, "This test only produces plots")
    def test_efficiency_catcher(self):
        setLogLevel(logging.DEBUG)

        plotter = Plotter(input_table=self.table)
        plotter.plot_efficiency()
        plotter.save_fig('test.png', efficiency=True)

    @unittest.skipIf(skip_plot, "This test only produces plots")
    def test_table(self):
        setLogLevel(logging.DEBUG)
        for target in self.targets:
            plotter = Plotter(input_table=self.targets[target].table)
            plotter.plot_table()
            plt.show()



from abc import ABCMeta
from abc import abstractmethod

from exorad.log.logger import Logger


class Task(Logger):
    """
    Base class for tasks.
    """

    __metaclass__ = ABCMeta

    _output = None
    _task_input = None
    _task_params = None

    @abstractmethod
    def __init__(self):
        """
        Class initialisation, needed to prepare the task inputs reader
        """
        pass

    @abstractmethod
    def execute(self):
        """
        Class execution. It runs on call and executes all the task actions returning the outputs.
        It requires the input with correct keywords
        """
        pass

    def __call__(self, **kwargs):
        self.set_log_name()
        self._task_input = kwargs
        self._validate_input_params()
        self._populate_empty_param()
        self.execute()
        return self.get_output()

    def _validate_input_params(self):
        for key in self._task_input.keys():
            if key not in self._task_params.keys():
                self.error("Unexpected Task input parameter: {}".format(key))
                raise ValueError

    def _populate_empty_param(self):
        for key in self._task_params.keys():
            if key not in self._task_input.keys():
                self._task_input[key] = self._task_params[key]["default"]

    def get_output(self):
        return self._output

    def set_output(self, product):
        self._output = product

    def get_task_param(self, paramName):
        return self._task_input[paramName]

    def addTaskParam(self, param_name, param_description, default=None):
        if self._task_params is None:
            self._task_params = {}
        self._task_params[param_name] = {
            "description": param_description,
            "default": default,
        }



import logging

import astropy.constants as const
import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

from exorad.models.signal import Signal
from exorad.utils.exolib import OmegaPix

logger = logging.getLogger("exorad.diffuse light")


def prepare(ch_table, ch_built_instr, description):
    logger.info("computing signal")
    wl_table = ch_table["Wavelength"]
    logger.debug("wl table : {}".format(wl_table))

    fnum_x = description["Fnum_x"]["value"].value
    if "Fnum_y" in description.keys():
        fnum_y = description["Fnum_y"]["value"].value
    else:
        fnum_y = None
    omega_pix = OmegaPix(fnum_x, fnum_y)
    logger.debug("omega pix : {}".format(omega_pix))
    A = (
        description["detector"]["delta_pix"]["value"]
        * description["detector"]["delta_pix"]["value"]
    ).to(u.m**2)
    qe = Signal(
        ch_built_instr["qe_data"]["wl_grid"]["value"]
        * u.Unit(ch_built_instr["qe_data"]["wl_grid"]["unit"]),
        ch_built_instr["qe_data"]["data"]["value"],
    )
    logger.debug("wl qe : {}".format(qe.wl_grid))

    transmission = Signal(
        ch_built_instr["transmission_data"]["wl_grid"]["value"]
        * u.Unit(ch_built_instr["transmission_data"]["wl_grid"]["unit"]),
        ch_built_instr["transmission_data"]["data"]["value"],
    )

    total_signal = np.zeros(wl_table.size) * u.count / u.s
    total_max_signal = np.zeros(wl_table.size) * u.count / u.s
    return (
        total_max_signal,
        total_signal,
        wl_table,
        A,
        qe,
        omega_pix,
        transmission,
    )


def convolve_with_slit(
    ch_description, ch_built_instr, A, ch_table, omega_pix, qe, radiance
):
    slit_width = ch_built_instr["slit_width"]
    wl_pix = ch_built_instr["wl_pix_center"]
    dwl_pic = ch_built_instr["pixel_bandwidth"]
    radiance.spectral_rebin(wl_pix)
    logger.debug("radiance : {}".format(radiance.data))

    qe_func = interp1d(
        qe.wl_grid,
        qe.data,
        assume_sorted=False,
        fill_value=0.0,
        bounds_error=False,
    )
    aomega = (
        omega_pix
        * A
        * qe_func(wl_pix)
        * (wl_pix / const.c / const.h).to(1.0 / u.W / u.s)
        * u.count
    )
    logger.debug("AOmega : {}".format(aomega))
    radiance.data *= aomega
    logger.debug("sed : {}".format(radiance.data))
    logger.debug("convolving with slit")
    slit_kernel = np.ones(
        int(
            slit_width
            / ch_description["detector"]["delta_pix"]["value"].to(u.um)
        )
    )
    signal_tmp = (
        np.convolve(radiance.data * dwl_pic, slit_kernel, "same")
    ).to(u.count / u.s)
    logger.debug("signal_tmp: {}".format(signal_tmp))
    idx = [
        np.logical_and(wl_pix > wlow, wl_pix <= whigh)
        for wlow, whigh in zip(
            ch_table["LeftBinEdge"], ch_table["RightBinEdge"]
        )
    ]
    signal = [signal_tmp[idx[k]].sum() for k in range(len(idx))]
    signal = [
        sig * w
        for sig, w in zip(
            signal, np.array(ch_built_instr["window_spatial_width"])
        )
    ]
    logger.debug("signal: {}".format(signal))
    try:
        max_signal_per_pix = [
            signal_tmp[idx[k]].max() for k in range(len(idx))
        ]
    except ValueError:
        logger.error("no max value in the array")
        raise
    return max_signal_per_pix, signal


def integrate_light(radiance, wl_qe, ch_built_instr):
    logger.debug("sed : {}".format(radiance.data))
    signal_tmp = np.trapz(radiance.data, x=wl_qe).to(u.count / u.s)
    # signal_tmp = (np.trapz(radiance.data[~np.isnan(radiance.data)], x=wl_qe[~np.isnan(radiance.data)])).to(
    #     u.count / u.s)
    signal = signal_tmp * np.asarray(ch_built_instr["window_size_px"])
    max_signal_per_pix = signal_tmp * np.ones_like(
        ch_built_instr["window_size_px"]
    )
    return max_signal_per_pix, signal



import unittest

from exorad.log import Logger


class LoggerTest(Logger):
    def __init__(self):
        super().__init__()
        self.info('info')
        self.debug('debug')
        self.warning('warning')
        self.critical('critical')
        self.error('error')


class LoadOptionsTest(unittest.TestCase):
    LoggerTest()

    def test_logs_messages(self):
        with self.assertLogs('exorad', level='DEBUG') as cm:
            LoggerTest()
            self.assertIn(
                "INFO:exorad.LoggerTest:info",
                cm.output)
            self.assertIn(
                "DEBUG:exorad.LoggerTest:debug",
                cm.output)
            self.assertIn(
                "WARNING:exorad.LoggerTest:warning",
                cm.output)
            self.assertIn(
                "CRITICAL:exorad.LoggerTest:critical",
                cm.output)
            self.assertIn(
                "ERROR:exorad.LoggerTest:error",
                cm.output)

    # def test_log_file(self):
    #     print('starts here')
    #
    #     fname = os.path.join(test_dir, 'exorad.log')
    #     try:
    #         os.remove(fname)
    #     except OSError:
    #         pass
    #     addLogFile(fname)
    #     LoggerTest()



from functools import lru_cache
from functools import wraps


@lru_cache(maxsize=2)
def nprocs():
    """Gets number of processes or returns 1 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 1 if MPI is not installed

    """
    try:
        from mpi4py import MPI
    except ImportError:
        return 1

    comm = MPI.COMM_WORLD

    return comm.Get_size()


@lru_cache(maxsize=2)
def get_rank():
    """Gets rank or returns 0 if mpi is not installed

    Returns
    -------
    int:
        Rank of process or 0 if MPI is not installed

    """

    try:
        from mpi4py import MPI
    except ImportError:
        return 0

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    return rank


def scatter(data, root=0):
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    if len(data) < nprocs():
        raise ValueError("not enough processes! work in progress")
    else:
        data = comm.scatter(data, root=root)
    return data


def gather(data, root=0):
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    data = comm.gather(data, root=root)
    return data


# def only_master_rank(f):
#     """
#     A decorator to ensure only the master
#     MPI rank can run it
#     """
#     @wraps(f)
#     def wrapper(*args, **kwargs):
#         if get_rank() == 0:
#             return f(*args, **kwargs)
#     return wrapper



import logging
import unittest

import astropy.units as u
import numpy as np

from exorad.log import setLogLevel
from exorad.models.noise import Noise
from exorad.models.signal import CountsPerSeconds
from exorad.models.signal import Sed
from exorad.models.signal import Signal

setLogLevel(logging.DEBUG)


class SignalTest(unittest.TestCase):

    def test_quantity_check(self):
        try:
            wl = np.linspace(0.1, 1, 10) * u.um
            data = np.random.random_sample((10, 10))
            time_grid = np.linspace(1, 5, 10) * u.hr
            Signal(wl_grid=wl, data=data, time_grid=time_grid)
            CountsPerSeconds(wl_grid=wl, data=data * u.Unit('ct/s'),
                             time_grid=time_grid)
            Noise(wl_grid=wl, data=data * u.hr ** 0.5, time_grid=time_grid)
            Sed(wl_grid=wl, data=data * u.W / u.m ** 2 / u.um,
                time_grid=time_grid)

            wl = np.linspace(0.1, 1, 10) * u.m
            time_grid = np.linspace(1, 5, 10) * u.s
            Signal(wl_grid=wl, data=data, time_grid=time_grid)

            wl = np.linspace(0.1, 1, 10)
            data = np.random.random_sample(10)
            Signal(wl_grid=wl, data=data)
            CountsPerSeconds(wl_grid=wl, data=data)
            Noise(wl_grid=wl, data=data)
            Sed(wl_grid=wl, data=data)

        except u.UnitConversionError:
            self.fail("Signal raised Exception unexpectedly!")

        with self.assertRaises(u.UnitConversionError):
            wl = np.linspace(0.1, 1, 10) * u.s
            data = np.random.random_sample(10)
            Signal(wl_grid=wl, data=data)

    def test_time_dependency(self):
        wl = np.linspace(0.1, 1, 10)
        data = np.random.random_sample((10, 10))
        time_grid = np.linspace(1, 5, 10)
        sig = Signal(wl_grid=wl, data=data, time_grid=time_grid)
        with self.assertRaises(NotImplementedError):
            sig.time_dependent
        fl = CountsPerSeconds(wl_grid=wl, data=data, time_grid=time_grid)
        with self.assertRaises(NotImplementedError):
            fl.time_dependent

        data = np.random.random_sample(10)
        sig = Signal(wl_grid=wl, data=data)
        self.assertFalse(sig.time_dependent)
        fl = CountsPerSeconds(wl_grid=wl, data=data)
        self.assertFalse(fl.time_dependent)

    def test_dimension_check(self):
        with self.assertRaises(ValueError):
            wl = np.linspace(0.1, 1, 10) * u.um
            data = np.random.random_sample((10, 2))
            Signal(wl_grid=wl, data=data)

            wl = np.linspace(0.1, 1, 10) * u.um
            data = np.random.random_sample(10)
            time_grid = np.linspace(1, 5, 10) * u.hr
            Signal(wl_grid=wl, data=data, time_grid=time_grid)

    def test_rebins(self):
        pass



import logging
from collections import OrderedDict

import numpy as np
from astropy import units as u

from exorad.models.signal import CustomSignal
from exorad.models.signal import Signal

logger = logging.getLogger("exorad.noise")


def frame_time(target, channel, out):
    """
    Given the channel and channel descriptions, populates the output table with saturation and frame times

    Parameters
    -----------
    channel: dict
        channel description
    target: Target
        Target to investigate
    out: QTable
        output table


    Returns
    --------
    QTable
        output table populated
    """
    name = channel["value"]
    max_signal_in_pix = target.table["MaxSignal_inPixel"][
        target.table["chName"] == name
    ]
    out["saturation_time"] = (
        channel["detector"]["well_depth"]["value"] / max_signal_in_pix
    )
    logger.debug("saturation time : {}".format(out["saturation_time"]))
    if "frame_time" in channel["detector"].keys():
        out["frameTime"] = channel["detector"]["frame_time"][
            "value"
        ] * np.ones(out["saturation_time"].size)
    else:
        out["frameTime"] = (
            channel["detector"]["f_well_depth"]["value"]
            * np.min(out["saturation_time"])
            * np.ones(out["saturation_time"].size)
        )
    logger.debug("frame time : {}".format(out["frameTime"]))
    return out


def multiaccum(channel, t_frame):
    """
    Given the channel and time frame, returns the multiaccum estimation for read and shot gain

    Parameters
    -----------
    channel: dict
        channel description
    t_frame: float
        frame time

    Returns
    --------
    float
        read noise gain
    float
        shot noise gain
    """
    nRead = np.floor(t_frame * channel["detector"]["freqNDR"]["value"])

    if "multiaccumM" in channel["detector"]:
        m = channel["detector"]["multiaccumM"]["value"]
        logger.debug("multiaccum activated: m = {}".format(m))
        tf = 0.0 * u.s
    else:
        m = 1
        tf = 0.0 * u.s

    if nRead < 2:
        nRead = 2.0  # Force to CDS in nRead < 2

    read_gain = 12.0 * (nRead - 1.0) / (nRead**2 + nRead) / m
    shot_gain = (
        6.0
        * (nRead**2 + 1.0)
        / (nRead**2 + nRead)
        / 5.0
        * (
            1
            - 5.0
            / 3.0
            * (m**2 - 1)
            / m
            / (nRead**2 + 1)
            * tf
            / (nRead - 1)
            / t_frame
        )
    )
    logger.debug("read noise gain: {}".format(read_gain))
    logger.debug("shot noise gain: {}".format(shot_gain))
    return read_gain, shot_gain


def photon_noise(table, channel, shot_gain, out):
    """
    Given the channel and channel descriptions, populates the output table with photon noises

    Parameters
    -----------
    channel: dict
        channel description
    target: Table
        Target table to investigate
    shot_gain: float
        multiaccum factor for photon noise
    out: QTable
        output table

    Returns
    --------
    QTable
        output table populated
    """
    name = channel["value"]

    signals = [key for key in table.keys() if "signal" in key]
    for key in signals:
        noise_key = "{}_noise".format(key)
        out[noise_key] = (
            np.sqrt(
                shot_gain
                * table[key][table["chName"] == name]
                * u.count
                / u.hr
            ).to(u.count / u.s)
            * u.hr**0.5
        )
        logger.debug("{} : {}".format(noise_key, out[noise_key]))
    return out


def photon_noise_variance(table, out):
    '''
    """
    Given the target table return the sum of photon variances

    Parameters
    -----------
    target: Table
        Target table to investigate
    out: QTable
        output table

    Returns
    --------
    Array
        sum of photon variances
    """
    '''
    signals = [key for key in table.keys() if "signal" in key]
    _photon_noise_variance = (
        np.zeros(out["frameTime"].size) * (u.count / u.s) ** 2 * u.hr
    )
    for key in signals:
        noise_key = "{}_noise".format(key)
        _photon_noise_variance += out[noise_key] * out[noise_key]
    return _photon_noise_variance


def add_custom_noise(custom, wl, out):
    if "data" in custom:
        col_name = [
            col for col in custom["data"].keys() if "Wavelength" not in col
        ]
        custom_noise = Signal(
            custom["data"]["Wavelength"], custom["data"][col_name]
        )
        custom_noise.spectral_rebin(wl)
        out["{}_noise".format(col_name)] = custom_noise.data
        logger.debug(
            "{} added as custom noise :{}".format(col_name, custom_noise.data)
        )
        out["total_noise"] = np.sqrt(
            out["total_noise"] * out["total_noise"]
            + custom_noise.data * custom_noise.data
        )

    if isinstance(custom, OrderedDict):
        for contrib in custom:
            custom_noise = (
                custom[contrib]["value"]
                * 1e-6
                * np.ones(wl.size)
                * u.hr**0.5
            )
            out[
                "{}_noise".format(custom[contrib]["name"]["value"])
            ] = custom_noise
            logger.debug(
                "{} added as custom noise :{}".format(
                    custom[contrib]["name"]["value"], custom_noise
                )
            )
            out["total_noise"] = np.sqrt(
                out["total_noise"] * out["total_noise"]
                + custom_noise * custom_noise
            )
    else:
        custom_noise = custom["value"] * 1e-6 * np.ones(wl.size) * u.hr**0.5
        out["{}_noise".format(custom["name"]["value"])] = custom_noise
        logger.debug(
            "{} added as custom noise :{}".format(
                custom["name"]["value"], custom_noise
            )
        )
        out["total_noise"] = np.sqrt(
            out["total_noise"] * out["total_noise"]
            + custom_noise * custom_noise
        )
    return out


class Noise(CustomSignal):
    """
    It's a Signal class with data having units of [hr^1/2]
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        super().__init__(wl_grid, data, u.hr**0.5, time_grid)



import numpy as np
from astropy import units as u
from scipy.interpolate import interp1d

from exorad.log import Logger
from exorad.models.utils import get_wl_col_name


class OpticalElement(Logger):
    """
    Handler for optical element

    Parameters
    -----------
    el: dict
        OpticalElement from the payload description file
    wl: array
        wavelength grid where to evaluate the element optical properties

    Attributes
    ----------
    name: str
        optical element name
    description: dict
        optical element input dict
    wl: array
        wavelength grid
    transmission: array
        optical element transmission estimated over the wl grid
    emissivity: array
        optical element emissivity
    temperature: float
        optical element temperature in K
    type: str
        optical element type
    position: str
        optical element position in the optic path

    """

    def __init__(self, description, wl):
        super().__init__()
        self.name = description["value"]
        self.debug(self.name)
        self.description = description
        self.wl = wl
        self.type = self.description["type"]["value"]
        self.transmission = self._get_transmission()
        self.emissivity = self._get_emissivity()
        self.temperature = self._get_temperature()
        self.position = self._get_position()

    def _get_position(self):
        if self.type == "detector box":
            return "detector"
        elif self.type == "optics box":
            return "optics box"
        else:
            return "path"

    def _get_temperature(self):
        if "temperature" in self.description:
            if hasattr(self.description["temperature"]["value"], "unit"):
                return self.description["temperature"]["value"].to(u.K)
            else:
                self.debug("Temperature assumed to be in K")
                return self.description["temperature"]["value"] * u.K
        else:
            return None

    def _get_transmission(self):
        """
        it return the transmission for the optical element.
        If a data file is indicated, the user should specify the column name to use.
        Such use allow the use to select if wants to use transmission of reflectivity for dicroics.
        If "use" is not indicated in the description file, the code will look for "Transmission" keyword in column name.
        """
        if "data" in self.description:
            self.debug("found transmission file")
            if "use" in self.description:
                colname = self.description["use"]["value"]
            else:
                try:
                    colname = "Transmission"
                    _ = self.description["data"][colname]
                except KeyError:
                    try:
                        colname = "Reflectivity"
                        _ = self.description["data"][colname]
                    except KeyError:
                        self.error(
                            "{} column name not found in transmission data file".format(
                                colname
                            )
                        )
                        raise KeyError(
                            "{} column name not found in transmission data file".format(
                                colname
                            )
                        )

            wl_colname = get_wl_col_name(
                self.description["data"], description=self.description
            )
            tr_func = interp1d(
                self.description["data"][wl_colname],
                self.description["data"][colname],
                assume_sorted=False,
                fill_value=0.0,
                bounds_error=False,
            )

            transmission = tr_func(self.wl)
        elif (
            "transmission" in self.description
            and "reflectivity" in self.description
        ):
            if "use" in self.description:
                key = self.description["use"]["value"]
                transmission = (
                    np.ones(self.wl.size) * self.description[key]["value"]
                )
            else:
                self.error(
                    "both transmission and reflectivity are included but use is not indicated"
                )
                raise KeyError(
                    "both transmission and reflectivity are included but use is not indicated"
                )

        elif (
            "transmission" in self.description
            or "reflectivity" in self.description
        ):
            try:
                transmission = (
                    np.ones(self.wl.size)
                    * self.description["transmission"]["value"]
                )
            except KeyError:
                transmission = (
                    np.ones(self.wl.size)
                    * self.description["reflectivity"]["value"]
                )
                self.debug(
                    "reflectivity keyword found for transmission for {}".format(
                        self.name
                    )
                )
            if "wl_min" in self.description:
                self.debug(
                    "found transmission inferior boundary : {}".format(
                        self.description["wl_min"]["value"]
                    )
                )
                idx = np.where(
                    self.wl
                    < self.description["wl_min"]["value"].to(self.wl.unit)
                )
                transmission[idx] = 0.0
            if "wl_max" in self.description:
                self.debug(
                    "found transmission superior boundary : {}".format(
                        self.description["wl_min"]["value"]
                    )
                )
                idx = np.where(
                    self.wl
                    > self.description["wl_max"]["value"].to(self.wl.unit)
                )
                transmission[idx] = 0.0

        else:
            transmission = np.ones(self.wl.size)
        self.debug("transmission : {}".format(transmission))
        return transmission

    def _get_emissivity(self):
        if "data" in self.description:
            self.debug("found emissivity file")

            wl_colname = get_wl_col_name(
                self.description["data"], description=self.description
            )

            em_colname = None
            for em_key in ["Emissivity", "emissivity"]:
                if em_key in self.description["data"].keys():
                    em_colname = em_key
            if em_colname is None:
                self.error(
                    "Emissivity column not found in transmission data file"
                )
                raise KeyError(
                    "Emissivity column not found in transmission data file"
                )

            em_func = interp1d(
                self.description["data"][wl_colname],
                self.description["data"][em_colname],
                assume_sorted=False,
                fill_value=0.0,
                bounds_error=False,
            )

            emissivity = em_func(self.wl)
        elif "emissivity" in self.description:
            emissivity = (
                np.ones(self.wl.size) * self.description["emissivity"]["value"]
            )
        elif self.type == "detector box":
            emissivity = np.ones(self.wl.size)
        else:
            emissivity = np.zeros(self.wl.size)
        self.debug("emissivity : {}".format(emissivity))
        return emissivity



import os
from copy import deepcopy

from ..models.targetlist import CSVTargetList
from ..models.targetlist import CSVTargetListMRS
from ..models.targetlist import QTableTargetList
from ..models.targetlist import XLXSTargetList
from .task import Task
from exorad.__version__ import __version__
from exorad.log import disableLogging
from exorad.log import enableLogging


class LoadTargetList(Task):
    """
    Loads target list from file

    Parameters
    ----------
    target_list: str
        target list file address. Supported formats are csv and xml


    Returns
    -------
    TargetList:
        return a list of Target class

    Examples
    --------
    >>> loadTargetList = LoadTargetList()
    >>> targets = loadTargetList(target_list='target_list/address')
    """

    def __init__(self):
        self.addTaskParam("target_list", "target list file name")

    def execute(self):
        target_list_file = self.get_task_param("target_list")
        self.info("target list file : {}".format(target_list_file))

        target_list_format = {
            ".xlsx": XLXSTargetList,
            ".csv": CSVTargetList,
        }

        try:
            ext = os.path.splitext(target_list_file)[1]
            self.info("target list file : {}".format(target_list_file))
            self.debug("target list format : {}".format(ext))
            target_klass = target_list_format[ext]
            tt = target_klass(target_list_file)
            if not tt.target:
                if ext == ".csv":
                    tt = CSVTargetListMRS(target_list_file)

        except KeyError:
            self.error("unsupported target list format: {}".format(ext))
            raise OSError("unsupported target list format: {}".format(ext))
        except TypeError:
            self.debug("target list is not a file. It is assumed to be QTable")
            tt = QTableTargetList(target_list_file)

        self.set_output(tt)


class PrepareTarget(Task):
    """
    Prepares the target output table over the channels to populate with light propagation

    Parameters
    ----------
    target: Target
        target to prepare
    channels : dict
        channel dictionary

    Returns
    -------
    Target:
        same target with table attribute populated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")
        self.addTaskParam("channels", "channels dictionary")

    def execute(self):
        from . import MergeChannelsOutput

        target = self.get_task_param("target")
        channels = self.get_task_param("channels")

        mergeChannelsOutput = MergeChannelsOutput()
        table = mergeChannelsOutput(channels=channels)
        table = self.add_metadata(table, target)
        target.table = table
        self.set_output(target)

    def add_metadata(self, table, target):
        table.meta["name"] = target.name
        table.meta["ExoRad version"] = __version__
        return table


class UpdateTargetTable(Task):
    """
    Updated the target output table over the channels to populate with light propagation

    Parameters
    ----------
    target: Target
        target to prepare
    table : QTable
        table to merge in the target table

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")
        self.addTaskParam("table", "table to merge in the target table")

    def execute(self):
        from astropy.table import hstack

        self.info("updating target table")
        target = self.get_task_param("target")
        target_table = target.table
        table = self.get_task_param("table")

        repeated_keys = [
            key for key in target_table.keys() if key in table.keys()
        ]
        if repeated_keys:
            self.debug(
                "the following keys are already in both table and will be replaced : {}".format(
                    repeated_keys
                )
            )
            target.table.remove_columns(repeated_keys)
        target.table = hstack([target_table, table])
        target.table.meta.update(table.meta)
        self.set_output(target)


class EstimateMaxSignal(Task):
    """
    Updated the target output table with maximum estimated signal in pix column

    Parameters
    ----------
    target: Target
        target to prepare

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")

    def execute(self):
        import numpy as np
        import astropy.units as u
        from astropy.table import QTable

        updateTargetTable = UpdateTargetTable()

        self.info("estimating max signal in pixel")
        target = self.get_task_param("target")
        table = target.table
        signals = [key for key in table.keys() if "Max" in key]
        max_signal = np.zeros(table["Wavelength"].size) * u.count / u.s
        for key in signals:
            max_signal += table[key]
        new_tab = QTable()
        new_tab["MaxSignal_inPixel"] = max_signal
        new_target = updateTargetTable(target=target, table=new_tab)
        self.set_output(new_target)


class ObserveTarget(Task):
    """
    Standard pipeline for target observation. It includes:
    1. PrepareTarget,
    2. EstimateForegrounds,
    3. PropagateForegroundLight
    4. LoadSource
    5. PropagateTargetLight
    6. EstimateNoise

    Parameters
    ----------
    target: Target
        target to prepare
    payload : dict
        payload description
    channels : dict
        channel dictionary
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")
        self.addTaskParam("payload", "payload description")
        self.addTaskParam("channels", "channel dictionary")
        self.addTaskParam(
            "wl_range", "wavelength range to investigate. (wl_min, wl_max)"
        )

    def execute(self):
        target = self.get_task_param("target")
        payload = self.get_task_param("payload")
        channels = self.get_task_param("channels")
        wl_min, wl_max = self.get_task_param("wl_range")

        from . import (
            PrepareTarget,
            PropagateTargetLight,
            EstimateForegrounds,
            PropagateForegroundLight,
            EstimateNoise,
            LoadSource,
        )

        prepareTarget = PrepareTarget()
        propagateTargetLight = PropagateTargetLight()
        loadSource = LoadSource()
        estimateForegrounds = EstimateForegrounds()
        propagateForegroundLight = PropagateForegroundLight()
        estimateNoise = EstimateNoise()

        target = prepareTarget(target=target, channels=channels)

        if "foreground" in payload["common"]:
            target = estimateForegrounds(
                foregrounds=payload["common"]["foreground"],
                target=target,
                wl_range=(wl_min, wl_max),
            )
        target = propagateForegroundLight(channels=channels, target=target)

        target, sed = loadSource(
            target=target,
            source=payload["common"]["sourceSpectrum"],
            wl_range=(wl_min, wl_max),
        )
        target = propagateTargetLight(channels=channels, target=target)

        target = estimateNoise(target=target, channels=channels)

        self.set_output(target)


def pipeline_to_dict(
    target,
    payload,
    channels,
    wl_range,
    plot,
    out_dir,
    debug,
):
    """This will be executed using concurrent futures"""
    from . import ObserveTarget
    from exorad.log.logger import root_logger

    observeTarget = ObserveTarget()

    root_logger.info("observing {}".format(target.name))
    if not debug:
        disableLogging()
    try:
        target = observeTarget(
            target=target,
            payload=payload,
            channels=channels,
            wl_range=wl_range,
        )
        enableLogging()
        outputDict = deepcopy(target)

        if plot:
            import matplotlib.pyplot as plt
            import matplotlib
            from exorad.utils.plotter import Plotter

            matplotlib.use("Agg")
            plotter = Plotter(input_table=target.table)
            plotter.plot_table()
            plotter.save_fig(
                os.path.join(out_dir, "{}.png".format(target.name))
            )
            plt.close()
        return target.name, outputDict
    except:
        enableLogging()
        root_logger.warning(
            "target {} skipped. Please check for previous error messages".format(
                target.name
            )
        )
        return None, None


class ObserveTargetlist(Task):
    """
    Standard pipeline to observe a full targetlist. It allows parallelization:

    Parameters
    ----------
    targets: Target
        targets to prepare
    payload : dict
        payload description
    channels : dict
        channel dictionary
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)
    plot: bool
        allow to save plots
    out_dir: str
        indicate the directory where to save plots
    n_thread: int
        number of threads
    debug: bool
        debug mode

    Returns
    -------
    dict:
        targets dict
    """

    def __init__(self):
        self.addTaskParam("targets", "targets to prepare")
        self.addTaskParam("payload", "payload description")
        self.addTaskParam("channels", "channel dictionary")
        self.addTaskParam(
            "wl_range", "wavelength range to investigate. (wl_min, wl_max)"
        )
        self.addTaskParam("plot", "allow to save plots")
        self.addTaskParam(
            "out_dir", "indicate the directory where to save plots"
        )
        self.addTaskParam("n_thread", "number of threads", 1)
        self.addTaskParam("debug", "debug mode", False)

    def execute(self):
        targets = self.get_task_param("targets")
        n_thread = self.get_task_param("n_thread")
        payload = self.get_task_param("payload")
        channels = self.get_task_param("channels")
        wl_range = self.get_task_param("wl_range")
        plot = self.get_task_param("plot")
        out_dir = self.get_task_param("out_dir")
        debug = self.get_task_param("debug")
        outputDict = {}

        if n_thread > 1:
            from concurrent.futures import (
                ProcessPoolExecutor,
                ThreadPoolExecutor,
            )
            from itertools import repeat

            # Does the distribution and chunking for you
            # Switch to ThreadPoolExecutor if you want to use python threading
            with ProcessPoolExecutor(max_workers=n_thread) as executor:
                for t_name, output in executor.map(
                    pipeline_to_dict,
                    targets,
                    repeat(payload),
                    repeat(channels),
                    repeat(wl_range),
                    repeat(plot),
                    repeat(out_dir),
                    repeat(debug),
                ):
                    outputDict[t_name] = output
        else:
            outputDict = {}
            for target in targets:
                t_name, output = pipeline_to_dict(
                    target, payload, channels, wl_range, plot, out_dir, debug
                )
                outputDict[t_name] = output

        self.set_output(outputDict)

    # def pipeline_to_dict(self,  target, outputDict):
    #     from . import ObserveTarget
    #     from exorad.utils.plotter import Plotter
    #     import matplotlib.pyplot as plt
    #     import matplotlib
    #     matplotlib.use('Agg')

    #     payload = self.get_task_param('payload')
    #     channels = self.get_task_param('channels')
    #     wl_range = self.get_task_param('wl_range')
    #     plot = self.get_task_param('plot')
    #     out_dir = self.get_task_param('out_dir')
    #     debug = self.get_task_param('debug')
    #     observeTarget = ObserveTarget()

    #     self.info('observing {}'.format(target.name))
    #     if not debug: disableLogging()
    #     try:
    #         target = observeTarget(target=target, payload=payload, channels=channels, wl_range=wl_range)
    #         enableLogging()
    #         outputDict[target.name] = deepcopy(target)

    #         if plot:
    #             plotter = Plotter(input_table=target.table)
    #             plotter.plot_table()
    #             plotter.save_fig(os.path.join(out_dir, '{}.png'.format(target.name)))
    #             plt.close()

    #     except:
    #         enableLogging()
    #         self.warning('target {} skipped. Please check for previous error messages'.format(target.name))



import os
import unittest

import numpy as np
from astropy import units as u
from inputs import test_dir

from exorad.utils.exolib import binnedPSF
from exorad.utils.exolib import paosPSF
from exorad.utils.exolib import plot_imac


class CreatePrfTest(unittest.TestCase):

    def test_prf(self):
        prf, pixel_rf, extent = binnedPSF(F_x=10, F_y=12,
                                          wl=3.0 * u.micron,
                                          delta_pix=18.0 * u.micron,
                                          plot=False)
        print(prf.shape)
        self.assertIsInstance(prf, np.ndarray)
        self.assertIsInstance(pixel_rf, np.ndarray)
        self.assertIsInstance(extent, tuple)

    def test_plot_prf(self):

        prf, pixel_rf, extent = binnedPSF(F_x=33.5, F_y=33.5 * 1.1/0.73,
                                          wl=[0.55] * u.micron,
                                          delta_pix=18.0 * u.micron,
                                          plot=False)

        _ = plot_imac(prf, extent, xlim=(-60, 60), ylim=(-60, 60))
        self.assertEqual(_, 0)


class LoadPAOSPhotTest(unittest.TestCase):
    paos_data = os.path.join(test_dir, 'hdf5', 'PAOS_Ariel_FGS-VISPhot.h5')
    parameters = {'grid_size': 512,
                  'wavelength': 0.55}

    def setUp(self):
        if not os.path.exists(self.paos_data):
            self.skipTest('skipped test due to .h5 file not found')

    def test_loader(self):

        print(self.parameters['wavelength'])

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)
        print('prf', prf.shape)
        print('extent', extent.shape)

    def test_shape(self):

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)

        self.assertEqual(prf.shape,
                         (self.parameters['grid_size'], self.parameters['grid_size']),
                         'incorrect psf shape')
        self.assertEqual(extent.shape,
                         (4,),
                         'incorrect extent shape')

    def test_plot_prf(self):

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)

        _ = plot_imac(prf, extent, xlim=(-60, 60), ylim=(-60, 60))
        self.assertEqual(_, 0)


class LoadPAOSSpecTest(unittest.TestCase):
    paos_data = os.path.join(test_dir, 'hdf5', 'PAOS_Ariel_FGS-NIRSpec.h5')
    parameters = {'grid_size': 512,
                  'wavelength': 1.1}

    def setUp(self):
        if not os.path.exists(self.paos_data):
            self.skipTest('skipped test due to .h5 file not found')

    def test_loader(self):

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)
        print('prf', prf.shape)
        print('extent', extent.shape)

    def test_loader_interp(self):

        prf, pixel_rf, extent = paosPSF(wl=[1.5] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)
        print('prf', prf.shape)
        print('extent', extent.shape)

    def test_shape(self):

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)

        self.assertEqual(prf.shape,
                         (self.parameters['grid_size'], self.parameters['grid_size']),
                         'incorrect psf shape')
        self.assertEqual(extent.shape,
                         (4,),
                         'incorrect extent shape')

    def test_plot_prf(self):

        prf, pixel_rf, extent = paosPSF(wl=[self.parameters['wavelength']] * u.micron,
                                        delta_pix=18.0 * u.micron,
                                        filename=self.paos_data)

        _ = plot_imac(prf, extent, xlim=(-60, 60), ylim=(-60, 60))
        self.assertEqual(_, 0)



import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from exorad.log.logger import Logger
from exorad.utils.exolib import rebin


class Signal(Logger):
    """
    Signal class is the main class for every wavelength and time dependent object.
    By default it assume wavelength in um, data dimensionless and time in hours.

    Parameters
    ----------
    wl_grid: quantity
        wavelength grid. Must have a single axes
    data: array
        data table. Can have 2 axes
    time_grid: quantity (optional)
        time grid. Must have a single axes

    Attributes
    ----------
    wl_grid: quantity(um)
        wavelength grid. Must have a single axes
    data: quantity(dimensionless)
        data table. Can have 2 axes
    time_grid: quantity(hr)
        time grid. Must have a single axes
    time_dependent: bool
        it tells if the signal is time dependent (True) or not (False)

    Examples
    ---------
    >>> import numpy as np
    >>> wl = np.linspace(0.1,1, 10)*u.um
    >>> data = np.random.random_sample((10,10))
    >>> time = np.linspace(1,5, 10)*u.hr
    >>> signal = Signal(wl_grid=wl, data=data, time_grid=time)
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        """
        class constructor

        Parameters
        ----------
        wl_grid: quantity
            wavelength grid. Must have a single axes
        data: array
            data table. Can have 2 axes
        time_grid: quantity (optional)
            time grid. Must have a single axes
        """

        self.set_log_name()
        self.wl_grid = self.check_quantities(wl_grid, "um")
        self.data = self.check_quantities(data, "")
        self.time_grid = self.check_quantities(time_grid, "hr")
        self.check_sizes()

    @property
    def time_dependent(self):
        """
        It tells if the Signal is time dependent
        Returns
        -------
        bool:
            True if Signal is time dependent.

        """
        if all(self.time_grid == [0] * u.hr):
            return False
        else:
            #            return True
            raise NotImplementedError

    def check_quantities(self, quantity, units):
        """
        It assures that the input data has the desired units, if not it converts the data.
        If the data are dimensionless it assumes that are reported with the correct units.

        Parameters
        ----------
        quantity: quantity
        units: units or string

        Returns
        -------
        quantity

        Raises
        ------
        UnitConversionError
            if it cannot convert the original units into the desired ones

        """
        if isinstance(quantity, u.Quantity):
            if quantity.unit != units:
                try:
                    data_converted = quantity.to(u.Unit(units))
                    self.debug("input data: {}".format(quantity))
                    self.debug(
                        "converted {} to {}".format(quantity.unit, units)
                    )
                    self.debug("converted data: {}".format(data_converted))
                    return data_converted
                except u.UnitConversionError:
                    self.error(
                        "Impossible to convert {} to {}".format(
                            quantity.unit, units
                        )
                    )
                    raise u.UnitConversionError

            else:
                return quantity
        else:
            # if units == '':
            #     self.debug('assumed dimensionless quantity as input'.format(units))
            # else:
            #     self.debug('assumed {} as input units'.format(units))
            data_converted = np.array(quantity) * u.Unit(units)
            return data_converted

    def check_sizes(self):
        """
        Assures that the class attributes (wl_grid, data and time_grid) have the correct dimensions

        Raises
        -------
        ValueError
            if there is a dimension mismatch
        """
        if self.wl_grid.ndim > 1:
            self.error("wavelength grid cannot have multiple axes")
            raise ValueError
        if self.time_grid.ndim > 1:
            self.error("time grid cannot have multiple axes")
            raise ValueError
        if self.data.ndim > 2:
            self.error("data cannot have more than 2 axes")
            raise ValueError
        if self.data.shape[0] != self.wl_grid.shape[0]:
            self.error("dimension mismatch between wavelength grid and data")
            raise ValueError
        if self.time_grid.shape[0] > 1 or self.data.ndim > 1:
            if self.data.shape[1] - self.time_grid.shape[0] != 0:
                self.error("dimension mismatch between time grid and data")
                raise ValueError

    def spectral_rebin(self, new_wl_grid):
        """
        rebins the signal to the new wavelength grid
        """
        self.data = rebin(new_wl_grid, self.wl_grid, self.data)[1]
        idx = np.where(np.isnan(self.data))
        self.data[idx] = 0.0
        self.wl_grid = new_wl_grid

    def temporal_rebin(self, new_time_grid):
        """
        rebins the signal to the new time grid
        """
        raise NotImplementedError

    def to_dict(self):
        """
        Converts the Signal class into standard python dict.
        """
        from exorad.utils.util import to_dict

        signal_dict = to_dict(self)
        return signal_dict

    def plot(self, fig=None, ax=None, yscale=None, xscale=None, label=None):
        """
        plots the signal and return figure and axis or add the plot to an existing figure and axis
        """
        if (fig == None) and (ax == None):
            fig, ax = plt.subplots(1, 1)
            ax.set_xlabel(r"Wavelength [${}$]".format(self.wl_grid.unit))
            ax.set_ylabel(r"${}$".format(self.data.unit))
        ax.plot(self.wl_grid, self.data, label=label)

        if yscale is not None:
            ax.set_yscale(yscale)
        if xscale is not None:
            ax.set_xscale(xscale)
        return fig, ax


class CustomSignal(Signal):
    """
    It's a Signal class with data having units custom units
    """

    def __init__(self, wl_grid, data, data_unit, time_grid=[0] * u.hr):
        self.set_log_name()
        self.wl_grid = self.check_quantities(wl_grid, "um")
        self.data = self.check_quantities(data, data_unit)
        self.time_grid = self.check_quantities(time_grid, "hr")
        self.check_sizes()


class Sed(CustomSignal):
    """
    It's a Signal class with data having units of [W m^-2 mu^-1]
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        super().__init__(wl_grid, data, u.W / u.m**2 / u.um, time_grid)


class Radiance(CustomSignal):
    """
    It's a Signal class with data having units of [W m^-2 mu^-1 sr^-1]
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        super().__init__(
            wl_grid, data, u.W / u.m**2 / u.um / u.sr, time_grid
        )


class CountsPerSeconds(CustomSignal):
    """
    It's a Signal class with data having units of [ct/s]
    """

    def __init__(self, wl_grid, data, time_grid=[0] * u.hr):
        super().__init__(wl_grid, data, "ct/s", time_grid)



def get_wl_col_name(data, description=None):
    wl_colname = None
    if description is not None and "wl_col_name" in description:
        wl_colname = description["wl_col_name"]["value"]
    else:
        for wl_key in ["Wavelength", "wavelength"]:
            if wl_key in data.keys():
                wl_colname = wl_key
    if wl_key is None:
        raise KeyError("Wavelength column not found in transmission data file")
    return wl_colname
